/*
   Copyright (C) 1998 T. Scott Dattalo

This file is part of gpsim.

gpsim is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

gpsim is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with gpasm; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

#include <iostream.h>
#include <stdio.h>


class invalid_file_register;   // Forward reference

#include "trace.h"
//#include "14bit-processors.h"

#ifndef __14_BIT_REGISTERS_H__
#define __14_BIT_REGISTERS_H__


class _14bit_processor;

#include "breakpoints.h"
//#include "stimuli.h"
class stimulus;  // forward reference
class IOPIN;
class source_stimulus;
class Stimulus_Node;
class PORTB;

//---------------------------------------------------------
// FSR register
//

class FSR : public sfr_register
{
public:

  unsigned int register_page_bits;   /* Only used by the 12-bit core to define
                                        the valid paging bits in the FSR. */

  void put(unsigned int new_value);
};


//---------------------------------------------------------
// Status register
//

class Status_register : public sfr_register
{
public:

#define STATUS_Z_BIT   2
#define STATUS_C_BIT   0
#define STATUS_DC_BIT  1
#define STATUS_PD_BIT  3
#define STATUS_TO_BIT  4
#define STATUS_Z       (1<<STATUS_Z_BIT)
#define STATUS_C       (1<<STATUS_C_BIT)
#define STATUS_DC      (1<<STATUS_DC_BIT)
#define STATUS_PD      (1<<STATUS_PD_BIT)
#define STATUS_TO      (1<<STATUS_TO_BIT)
#define BREAK_Z_ACCESS 2
#define BREAK_Z_WRITE  1

#define RP_MASK        0x20

  unsigned int break_on_z,break_on_c;

  Status_register(void);

  inline void put(unsigned int new_value);

  inline unsigned int get(void)
    {
      if(break_point) 
	bp.check_read_break(this);

      trace.register_read(address,value);
      return(value);
    }

  // Special member function to control just the Z bit

       inline void put_Z(unsigned int new_z)
    {
      if(break_on_z) 
	bp.check_write_break(this);

      value = (value & ~STATUS_Z) | ((new_z) ? STATUS_Z : 0);

      trace.register_write(address,value);
    }

  inline unsigned int get_Z(void)
    {
      if(break_on_z) 
	bp.check_read_break(this);

      trace.register_read(address,value);
      return( ( (value & STATUS_Z) == 0) ? 0 : 1);
    }


  // Special member function to control just the C bit
       void put_C(unsigned int new_c)
    {
      if(break_on_c) 
	bp.check_write_break(this);

      value = (value & ~STATUS_C) | ((new_c) ? STATUS_C : 0);
      trace.register_write(address,value);
    }

  unsigned int get_C(void)
    {
      if(break_on_c) 
	bp.check_read_break(this);

      trace.register_read(address,value);
      return( ( (value & STATUS_C) == 0) ? 0 : 1);
    }

  // Special member function to set Z, C, and DC

       inline void put_Z_C_DC(unsigned int new_value, unsigned int src1, unsigned int src2)
    {
      if(break_point) 
	bp.check_write_break(this);

      value = (value & ~ (STATUS_Z | STATUS_C | STATUS_DC)) |  
	((new_value & 0xff)   ? 0 : STATUS_Z)   |
	((new_value & 0x100)  ? STATUS_C : 0)   |
	(((new_value ^ src1 ^ src2)&0x10) ? STATUS_DC : 0);

      trace.register_write(address,value);
    }

  inline void put_Z_C_DC_for_sub(unsigned int new_value, unsigned int src1, unsigned int src2)
    {
      if(break_point) 
	bp.check_write_break(this);

      value = (value & ~ (STATUS_Z | STATUS_C | STATUS_DC)) |  
	((new_value & 0xff)   ? 0 : STATUS_Z)   |
	((new_value & 0x100)  ? 0 : STATUS_C)   |
	(((new_value ^ src1 ^ src2)&0x10) ? 0 : STATUS_DC);

      trace.register_write(address,value);
    }

  inline void put_PD(unsigned int new_pd)
    {
      if(break_point) 
	bp.check_write_break(this);

      value = (value & ~STATUS_PD) | ((new_pd) ? STATUS_PD : 0);

      trace.register_write(address,value);
    }

  inline unsigned int get_PD(void)
    {
      if(break_point) 
	bp.check_read_break(this);

      trace.register_read(address,value);
      return( ( (value & STATUS_PD) == 0) ? 0 : 1);
    }

  inline void put_TO(unsigned int new_to)
    {
      if(break_point) 
	bp.check_write_break(this);

      value = (value & ~STATUS_TO) | ((new_to) ? STATUS_TO : 0);

      trace.register_write(address,value);
    }

  inline unsigned int get_TO(void)
    {
      if(break_point) 
	bp.check_read_break(this);

      trace.register_read(address,value);
      return( ( (value & STATUS_TO) == 0) ? 0 : 1);
    }



};


//---------------------------------------------------------
// Program Counter
//

class Program_Counter
{
public:
  pic_processor *cpu;
  unsigned int value;              /* pc's current value */
  unsigned int memory_size_mask; 

  Program_Counter(void);
  void increment(void);
  void skip(void);
  void jump(unsigned int new_value);
  void computed_goto(unsigned int new_value);
  void new_address(unsigned int new_value);
  void reset(void)
    { 
      value = 0;
      trace.program_counter(value);
    };

  unsigned int get_next(void);
};


//---------------------------------------------------------
// Stack
//

class Stack
{
public:
  unsigned int contents[8];        /* the stack array */ 
  int pointer;                     /* the stack pointer */
  unsigned int stack_mask;         /* 1 for 12bit, 7 for 14bit */
  bool stack_warnings_flag;        /* Should over/under flow warnings be printed? */

  Stack(void);
  void push(unsigned int);
  unsigned int pop(void);
  void reset(void) {pointer = 0;};  // %%% FIX ME %%% reset may need to change 
  // because I'm not sure how the stack is affected by a reset.
};


//---------------------------------------------------------
// Cycle Counter
//
// This not really a real PIC register. The cycle counter class
// is used to coordinate the timing between the different peripherals
// within a PIC and in some cases, the timing between several simulated
// PICs.

class Cycle_Counter_breakpoint_list
{
  // This class is used to implement a very simple single-linked-list.
  // See Cycle_Counter::set_break for more details on how it is used.

public:
  Cycle_Counter_breakpoint_list *next;
  unsigned int break_value;
  unsigned int breakpoint_number;  // sytem bp (this is used if the user set a bp.
				   // It tells which system bp needs to be cleared
                                   // when the cycle break is hit. If a peripheral set
                                   // a cycle break, then there is no corresponding 
                                   // system break.)
  BreakCallBack *f;
};

class Cycle_Counter
{
public:

#define BREAK_ARRAY_SIZE  32
#define BREAK_ARRAY_MASK  (BREAK_ARRAY_SIZE -1)

  //_14bit_processor *cpu;           // The cpu to which this counter belongs.
  unsigned int value;              // Current value of the cycle counter.
  unsigned int break_on_this;      // Break if the cycle counter reaches this value.

  Cycle_Counter_breakpoint_list
    active,     // Head of the active breakpoint linked list
    inactive;   // Head of the inactive one.

  Cycle_Counter(void);
  void preset(unsigned int new_value);     // not used currently.

  /*
    increment - This inline member function is called once or twice for every simulated
    instruction. Its purpose is to increment the cycle counter using roll over arithmetic.
    If there's a breakpoint set on the new value of the cycle counter then the simulation
    is either stopped or a callback function is invoked. In either case, the break point is
    cleared.
   */
						
  inline void increment(void)
    {
 
      // Increment the current cycle
      // we have a break point set here

	   if(++value == break_on_this)
	     {
	       // There's a break point set on this cycle. If there's a callback function, then call
	       // it other wise halt execution by setting the global break flag.

		    while(value == break_on_this)   // Loop in case there are multiple breaks
		      {
			if(active.next->f != NULL)
			  active.next->f->callback();
			else
			  bp.check_cycle_break(active.next->breakpoint_number);

			clear_current_break();
		      }
	     }

      trace.cycle_counter(value);
    }



  bool set_break(unsigned int future_cycle, BreakCallBack *f=NULL, unsigned int abp = MAX_BREAKPOINTS);
  bool reassign_break(unsigned int old_cycle,unsigned int future_cycle, BreakCallBack *f=NULL);
  void clear_current_break(void);
  void dump_breakpoints(void);

};


//---------------------------------------------------------
// Watch Dog Timer
//

class WDT : public BreakCallBack
{
public:
  pic_processor *cpu;           // The cpu to which this wdt belongs.

  unsigned int
    value,
    prescale,
    future_cycle,
    break_point;

  double timeout;
  bool   wdte;
  bool   warned;

  void put(unsigned int new_value);
  void initialize(bool enable, double _timeout);
  void clear(void);
  virtual void callback(void);
};

//---------------------------------------------------------
// W register

class WREG : public sfr_register
{
public:

  void put(unsigned int new_value);
  unsigned int get(void);
  WREG(void);
};

//---------------------------------------------------------
// IOPORT

class IOPORT_TRIS;  // forward reference

class IOPORT : public sfr_register
{
public:

#define IOPINS 8

  IOPIN  *pins[IOPINS];

  IOPORT_TRIS * tris;

  unsigned int 
    valid_iopins,   // A mask that for those ports that don't have all 8 io bits.
    stimulus_mask,  // A mask indicating which io bits have a stimulus.
    internal_latch; // 

  void put(unsigned int new_value);
  virtual void setbit(unsigned int bit_number, bool new_value);
  int get_bit(unsigned int bit_number);
  unsigned int get(void);
  void attach_stimulus(stimulus *new_stim, unsigned int bit_position);
  int update_stimuli(void);
  void attach_iopin(IOPIN * new_pin, unsigned int bit_position);
  void attach_node(Stimulus_Node *new_node, unsigned int bit_position);
  void update_pin_directions(unsigned int new_tris);

  IOPORT(void);

};

//---------------------------------------------------------
// IOPORT

class IOPORT_TRIS : public sfr_register
{
public:
  IOPORT * port;

  unsigned int 
    valid_iopins;   // A mask that for those ports that don't have all 8 io bits.

  void put(unsigned int new_value);
  unsigned int get(void);
  IOPORT_TRIS(void);

};


//---------------------------------------------------------
// TMR0 - Timer
class TMR0 : public sfr_register, public BreakCallBack
{
public:
  unsigned int 
    synchronized_cycle,
    prescale,
    prescale_counter,
    last_cycle,
    future_cycle;


  virtual void callback(void);

  TMR0(void);

  void put(unsigned int new_value);
  unsigned int get(void);
  unsigned int get_value(void);
  void start(void);
  void increment(void);   // Used when tmr0 is attached to an external clock
  void new_prescale(void);
  void new_clock_source(void);

};

//---------------------------------------------------------
// INTCON - Interrupt control register

class INTCON : public sfr_register
{
public:

const unsigned int
  RBIF = 1<<0,
  INTF = 1<<1,
  T0IF = 1<<2,
  RBIE = 1<<3,
  INTE = 1<<4,
  T0IE = 1<<5,
  XXIE = 1<<6,    // Processor dependent
  GIE  = 1<<7;



  INTCON(void);
  void set_T0IF(void);

  /*
  // Bit 6 of intcon depends on the processor that's being simulated, 
  // This generic function will get called whenever interrupt flag upon which bit 6 enables
  //  becomes true. (e.g. for the c84, this routine is called when EEIF goes high.)
  */
  inline void peripheral_interrupt(void)
    {
      if(  (value & (GIE | XXIE)) == (GIE | XXIE) ) bp.set_interrupt();
    };

  inline void set_gie(void)
    {
      value |= GIE;
      put(value);
    }

  inline void set_rbif(void)
    {
      put(get() | RBIF);
    }

  inline void set_intf(void)
    {
      put(get() | INTF);
    }

  inline void set_t0if(void)
    {
      put(get() | T0IF);
    }

  inline void set_rbie(void)
    {
      put(get() | RBIE);
    }

  inline void set_inte(void)
    {
      put(get() | INTE);
    }

  inline void set_t0ie(void)
    {
      put(get() | T0IE);
    }

  inline void clear_gie(void)
    {
      put(get() & ~GIE);
    }

  virtual bool check_peripheral_interrupt(void)
    {
      return 0;
    }

  void put(unsigned int new_value)
    {

//      if(break_point) 
//	bp.check_write_break(this);


      value = new_value;
      trace.register_write(address,value);

  // Now let's see if there's a pending interrupt
  // The INTCON bits are:
  // GIE | ---- | TOIE | INTE | RBIE | TOIF | INTF | RBIF
  // There are 3 sources for interrupts, TMR0, RB0/INTF
  // and RBIF (RB7:RB4 change). If the corresponding interrupt
  // flag is set AND the corresponding interrupt enable bit
  // is set AND global interrupts (GIE) are enabled, THEN
  // there's an interrupt pending.
  // note: bit6 is not handled here because it is processor
  // dependent (e.g. EEIE for x84 and ADIE for x7x).

      if(value & GIE)
        {

          if( (((value>>3)&value) & (T0IF | INTF | RBIF)) )
            {
              trace.interrupt();
              bp.set_interrupt();
            }
          else if(value & XXIE)
            {
              if(check_peripheral_interrupt())
                {
                  trace.interrupt();
                  bp.set_interrupt();
                }
            }
        }
    }

};
//---------------------------------------------------------
// INDF

class INDF : public sfr_register
{
public:

//  _14bit_processor *cpu;

  INDF(void);
  void put(unsigned int new_value);
  unsigned int get(void);
};

//---------------------------------------------------------
// OPTION_REG - 

class OPTION_REG : public sfr_register
{
public:
// const unsigned int
enum
  {
    PS0    = 1<<0,
    PS1    = 1<<1,
    PS2    = 1<<2,
    PSA    = 1<<3,
    T0SE   = 1<<4,
    T0CS   = 1<<5,
    BIT6   = 1<<6,
    BIT7   = 1<<7
  };

  unsigned int prescale;

//  PORTB *portb;  

  OPTION_REG(void);

  inline unsigned int get_prescale(void)
    {
      return value & (PS0 | PS1 | PS2);
    }

  inline unsigned int get_psa(void)
    {
      return value & PSA;
    }

  inline unsigned int get_t0cs(void)
    {
      return value & T0CS;
    }

  inline unsigned int get_t0se(void)
    {
      return value & T0SE;
    }

  void put(unsigned int new_value);

};


//---------------------------------------------------------
// PCL - Program Counter Low
//

class PCL : public sfr_register
{
public:
//  _14bit_processor *cpu;

  void put(unsigned int new_value);
  unsigned int get(void);

  PCL(void);
};

//---------------------------------------------------------
// PCLATH - Program Counter Latch High
//

class PCLATH : public sfr_register
{
public:
  void put(unsigned int new_value);
  unsigned int get(void);

  PCLATH(void);
};

class EEPROM;    // forward reference to the EE prom peripheral
class EECON2;    // forward reference for EECON1

//---------------------------------------------------------
// EECON1 - EE control register 1
//

class EECON1 : public sfr_register
{
public:
const unsigned int RD    = (1<<0);
const unsigned int WR    = (1<<1);
const unsigned int WREN  = (1<<2);
const unsigned int WRERR = (1<<3);
const unsigned int EEIF  = (1<<4);

const unsigned int VALID_BITS = (RD | WR | WREN | WRERR | EEIF);

  EEPROM *eeprom;

  void put(unsigned int new_value);
  unsigned int get(void);

  EECON1(void);
};

//
// EECON2 - EE control register 2
//

class EECON2 : public sfr_register
{
public:

enum EE_STATES
{
  NOT_READY,
  HAVE_0x55,
  READY,
  UNARMED
} eestate;

  EEPROM *eeprom;

  void put(unsigned int new_value);
  unsigned int get(void);
  void ee_reset(void) { eestate = NOT_READY;};

  EECON2(void);
};

//
// EEDATA - EE data register
//

class EEDATA : public sfr_register
{
public:

  EEPROM *eeprom;

  void put(unsigned int new_value);
  unsigned int get(void);

  EEDATA(void);
};

//
// EEADR - EE address register
//

class EEADR : public sfr_register
{
public:

  EEPROM *eeprom;

  void put(unsigned int new_value);
  unsigned int get(void);

  EEADR(void);
};


class EEPROM :  public BreakCallBack
{
public:

#define  EPROM_WRITE_TIME  20

  char * name_str;
  _14bit_processor *cpu;

  EECON1 *eecon1;            // The EEPROM consists of 4 control registers
  EECON2 *eecon2;
  EEDATA *eedata;
  EEADR  *eeadr;

  file_register **rom;       //  and the data area.
  unsigned int rom_size;
  unsigned int wr_adr,wr_data;  // latched adr and data for eewrites.
  unsigned int abp;             // break point number that's set during eewrites

  EEPROM(void);
  void reset(RESET_TYPE);
  virtual void callback(void);
  void start_write(void);
  void initialize(unsigned int new_rom_size,EECON1 *con1, EECON2 *con2, EEDATA *data, EEADR *adr);
  void dump(void);

};
#endif
